@@ -3,12 +3,12 @@ widgets.py
 
 Common GUI widgets that are re-used across plottr.
 """
-from numpy import rint
 from typing import Union, List, Tuple, Optional, Type, Sequence, Dict, Any, Type
 
 from .tools import dictToTreeWidgetItems, dpiScalingFactor
 from plottr import QtGui, QtCore, Flowchart, QtWidgets, Signal, Slot
-from plottr.node import Node, linearFlowchart
+from plottr.node import Node, linearFlowchart, NodeWidget, updateOption
+from plottr.node.node import updateGuiQuietly, emitGuiUpdate
 from ..plot import PlotNode, PlotWidgetContainer, PlotWidget
 from .. import config_entry as getcfg
 
@@ -305,3 +305,183 @@ class Collapsible(QtWidgets.QWidget):
         else:
             self.widget.setVisible(False)
             self.btn.setText(self.collapsedTitle)
+
+
+class DimensionCombo(QtWidgets.QComboBox):
+    """A Combo Box that allows selection of a single data dimension.
+    This widget is designed to be used in a node widget.
+
+    Which type of dimensions are available for selection is set through the
+    ``dimensionType`` option when creating the instance.
+
+    The widget can be linked to a node using the :meth:`.connectNode` method.
+    After linking, the available options will be populated whenever the data in
+    the node changes.
+    """
+
+    #: Signal(str)
+    #: emitted when the user selects a dimension.
+    dimensionSelected = Signal(str)
+
+    def __init__(self, parent: Optional[QtWidgets.QWidget] = None,
+                 dimensionType: str = 'axes') -> None:
+        """Constructor.
+
+        :param parent: parent widget
+        :param dimensionType: one of `axes`, `dependents` or `all`.
+        """
+        super().__init__(parent)
+
+        self.node: Optional[Node] = None
+        self.dimensionType = dimensionType
+
+        self.clear()
+        self.entries = ['None']
+        for e in self.entries:
+            self.addItem(e)
+
+        self.currentTextChanged.connect(self.signalDimensionSelection)
+
+    def connectNode(self, node: Optional[Node] = None) -> None:
+        """Connect a node. will result in populating the combo box options
+        based on dimensions available in the node data.
+
+        :param node: instance of :class:`.Node`
+        """
+        if node is None:
+            raise RuntimeError
+        self.node = node
+        if self.dimensionType == 'axes':
+            self.node.dataAxesChanged.connect(self.setDimensions)
+        elif self.dimensionType == 'dependents':
+            self.node.dataDependentsChanged.connect(self.setDimensions)
+        else:
+            self.node.dataFieldsChanged.connect(self.setDimensions)
+
+    @updateGuiQuietly
+    def setDimensions(self, dims: Sequence[str]) -> None:
+        """Set the dimensions that are available for selection.
+
+        :param dims: list of dimensions, as strings.
+        :return: ``None``
+        """
+        self.clear()
+        allDims = self.entries + list(dims)
+        for d in allDims:
+            self.addItem(d)
+
+    @Slot(str)
+    @emitGuiUpdate('dimensionSelected')
+    def signalDimensionSelection(self, val: str) -> str:
+        return val
+
+
+class DimensionSelector(FormLayoutWrapper):
+    """A widget that allows the user to select a dimension from a dataset
+    via a combobox.
+
+    Contains a label and a :class:`.DimensionCombo`."""
+
+    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
+        super().__init__(
+            parent=parent,
+            elements=[('Dimension', DimensionCombo(dimensionType='all'))],
+        )
+        self.combo = self.elements['Dimension']
+
+
+class DependentSelector(FormLayoutWrapper):
+    """A widget that allows the user to select a dependent dimension from a dataset
+    via a combobox.
+
+    Contains a label and a :class:`.DimensionCombo`."""
+
+    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
+        super().__init__(
+            parent=parent,
+            elements=[('Dependent', DimensionCombo(dimensionType='dependents'))],
+        )
+        self.combo = self.elements['Dependent']
+
+
+class AxisSelector(FormLayoutWrapper):
+    """A widget that allows the user to select an axis dimension from a dataset
+    via a combobox.
+
+    Contains a label and a :class:`.DimensionCombo`."""
+
+    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
+        super().__init__(
+            parent=parent,
+            elements=[('Axis', DimensionCombo(dimensionType='axes'))],
+        )
+        self.combo = self.elements['Axis']
+
+
+class MultiDimensionSelector(QtWidgets.QListWidget):
+    """A simple list widget that allows selection of multiple data dimensions."""
+
+    #: signal (List[str]) that is emitted when the selection is modified.
+    dimensionSelectionMade = Signal(list)
+
+    def __init__(self, parent: Optional[QtWidgets.QWidget] = None,
+                 dimensionType: str = 'all') -> None:
+        """Constructor.
+
+        :param parent: parent widget.
+        :param dimensionType: one of ``all``, ``axes``, or ``dependents``.
+        """
+        super().__init__(parent)
+
+        self.node: Optional[Node] = None
+        self.dimensionType = dimensionType
+
+        self.setSelectionMode(self.MultiSelection)
+        self.itemSelectionChanged.connect(self.emitSelection)
+
+    def setDimensions(self, dimensions: List[str]) -> None:
+        """set the available dimensions.
+
+        :param dimensions: list of dimension names.
+        """
+        self.clear()
+        self.addItems(dimensions)
+
+    def getSelected(self) -> List[str]:
+        """Get selected dimensions.
+
+        :return: List of dimensions (as strings).
+        """
+        selectedItems = self.selectedItems()
+        return [s.text() for s in selectedItems]
+
+    def setSelected(self, selected: List[str]) -> None:
+        """Set dimension selection.
+
+        :param selected: List of dimensions to be selected.
+        """
+        for i in range(self.count()):
+            item = self.item(i)
+            if item.text() in selected:
+                item.setSelected(True)
+            else:
+                item.setSelected(False)
+
+    def emitSelection(self) -> None:
+        self.dimensionSelectionMade.emit(self.getSelected())
+
+    def connectNode(self, node: Optional[Node] = None) -> None:
+        """Connect a node. Will result in populating the available options
+        based on dimensions available in the node data.
+
+        :param node: instance of :class:`.Node`
+        """
+        if node is None:
+            raise RuntimeError
+        self.node = node
+        if self.dimensionType == 'axes':
+            self.node.dataAxesChanged.connect(self.setDimensions)
+        elif self.dimensionType == 'dependents':
+            self.node.dataDependentsChanged.connect(self.setDimensions)
+        else:
+            self.node.dataFieldsChanged.connect(self.setDimensions)
